#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import unicode_literals

import argparse
import logging
import signal
import sys
from queue import Queue

import psutil
import yaml
from twisted.internet import inotify
from twisted.internet import reactor
from twisted.python import filepath

log = logging.getLogger('tron.recover_batch')


class StatusFileWatcher(object):
    """
    Watches the status file produced by action runners
    """

    def __init__(self, to_watch, callback):
        notifier = inotify.INotify()
        notifier.startReading()
        notifier.watch(filepath.FilePath(to_watch), callbacks=[callback])


def parse_args():
    parser = argparse.ArgumentParser(
        description='Check if a action runner has exited; wait otherwise',
    )
    parser.add_argument('filepath', )
    return parser.parse_args()


def notify(notify_queue, ignored, filepath, mask):
    with open(filepath.path) as f:
        last_entry = f.readlines()[-1]
        x = yaml.load(last_entry)
        return_code = x.get('return_code')
        if return_code is not None:
            reactor.stop()
            notify_queue.put(return_code)


def get_key_from_last_line(filepath, key):
    with open(filepath) as f:
        lines = f.readlines()
        if lines:
            content = yaml.load(lines[-1])
            return content.get(key)
        return None


if __name__ == "__main__":
    args = parse_args()
    existing_return_code = get_key_from_last_line(args.filepath, 'return_code')
    if existing_return_code is not None:
        sys.exit(existing_return_code)

    runner_pid = get_key_from_last_line(args.filepath, 'runner_pid')
    if not psutil.pid_exists(runner_pid):
        log.warning(
            "action_runner pid %d no longer running; unable to recover batch" %
            runner_pid
        )
        #TODO: should we kill the process here?
        sys.exit(1)

    exit_code_queue = Queue()
    watcher = StatusFileWatcher(
        args.filepath,
        lambda *args, **kwargs: notify(exit_code_queue, *args, **kwargs)
    )
    reactor.run()
    exit_code = exit_code_queue.get()
    assert type(exit_code) == int
    if exit_code < 0:
        # from the subprocess docs on the return code of a process:
        # "A negative value -N indicates that the child was terminated by signal N (POSIX only)."
        # We should always exit with a positive code, so we take the absolute value of the returncode
        log.warning(
            f'action run killed by signal {signal.Signals(abs(exit_code)).name}'
        )
        exit_code = abs(exit_code)
    sys.exit(exit_code)



from __future__ import absolute_import
from __future__ import unicode_literals

import logging

from tron import node
from tron.config.schema import CLEANUP_ACTION_NAME
from tron.utils import maybe_decode

log = logging.getLogger(__name__)


class Action(object):
    """A configurable data object for an Action."""

    equality_attributes = [
        'name', 'command', 'node_pool', 'is_cleanup', 'retries',
        'expected_runtime', 'executor', 'cpus', 'mem', 'constraints',
        'docker_image', 'docker_parameters', 'env', 'extra_volumes',
        'mesos_address', 'retries_delay'
    ]

    def __init__(
        self,
        name,
        command,
        node_pool,
        required_actions=None,
        dependent_actions=None,
        retries=None,
        retries_delay=None,
        expected_runtime=None,
        executor=None,
        cpus=None,
        mem=None,
        constraints=None,
        docker_image=None,
        docker_parameters=None,
        env=None,
        extra_volumes=None,
        mesos_address=None,
    ):
        self.name = maybe_decode(name)
        self.command = command
        self.node_pool = node_pool
        self.retries = retries
        self.retries_delay = retries_delay
        self.required_actions = required_actions or []
        self.dependent_actions = dependent_actions or []
        self.expected_runtime = expected_runtime
        self.executor = executor
        self.cpus = cpus
        self.mem = mem
        self.constraints = constraints or []
        self.docker_image = docker_image
        self.docker_parameters = docker_parameters or []
        self.env = env or {}
        self.extra_volumes = extra_volumes or []
        self.mesos_address = mesos_address

    @property
    def is_cleanup(self):
        return self.name == CLEANUP_ACTION_NAME

    @classmethod
    def from_config(cls, config):
        """Factory method for creating a new Action."""
        node_repo = node.NodePoolRepository.get_instance()

        # Only convert config values if they are not None.
        constraints = config.constraints
        if constraints:
            constraints = [[c.attribute, c.operator, c.value]
                           for c in constraints]
        docker_parameters = config.docker_parameters
        if docker_parameters:
            docker_parameters = [c._asdict() for c in docker_parameters]
        extra_volumes = config.extra_volumes
        if extra_volumes:
            extra_volumes = [c._asdict() for c in extra_volumes]

        return cls(
            name=config.name,
            command=config.command,
            node_pool=node_repo.get_by_name(config.node),
            retries=config.retries,
            retries_delay=config.retries_delay,
            expected_runtime=config.expected_runtime,
            executor=config.executor,
            cpus=config.cpus,
            mem=config.mem,
            constraints=constraints,
            docker_image=config.docker_image,
            docker_parameters=docker_parameters,
            env=config.env,
            extra_volumes=extra_volumes,
            mesos_address=config.mesos_address,
        )

    def __eq__(self, other):
        attributes_match = all(
            getattr(self, attr, None) == getattr(other, attr, None)
            for attr in self.equality_attributes
        )
        return attributes_match and all(
            self_act == other_act for (
                self_act,
                other_act,
            ) in zip(self.required_actions, other.required_actions)
        )

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.name)



from __future__ import absolute_import
from __future__ import unicode_literals

import mock
from testify import assert_equal
from testify import run
from testify import setup
from testify import TestCase

from tron import node
from tron.config.schema import ConfigAction
from tron.config.schema import ConfigConstraint
from tron.config.schema import ConfigParameter
from tron.config.schema import ConfigVolume
from tron.core import action


class TestAction(TestCase):
    @setup
    def setup_action(self):
        self.node_pool = mock.create_autospec(node.NodePool)
        self.action = action.Action("my_action", "doit", self.node_pool)

    def test_from_config_full(self):
        config = ConfigAction(
            name="ted",
            command="do something",
            node="first",
            executor="ssh",
            cpus=1,
            mem=100,
            constraints=[
                ConfigConstraint(
                    attribute='pool',
                    operator='LIKE',
                    value='default',
                ),
            ],
            docker_image='fake-docker.com:400/image',
            docker_parameters=[
                ConfigParameter(
                    key='test',
                    value=123,
                ),
            ],
            env={'TESTING': 'true'},
            extra_volumes=[
                ConfigVolume(
                    host_path='/tmp',
                    container_path='/nail/tmp',
                    mode='RO',
                ),
            ],
            mesos_address='fake-mesos-master.com',
        )
        new_action = action.Action.from_config(config)
        assert_equal(new_action.name, config.name)
        assert_equal(new_action.command, config.command)
        assert_equal(new_action.node_pool, None)
        assert_equal(new_action.required_actions, [])
        assert_equal(new_action.executor, config.executor)
        assert_equal(new_action.cpus, config.cpus)
        assert_equal(new_action.mem, config.mem)
        assert_equal(new_action.constraints, [['pool', 'LIKE', 'default']])
        assert_equal(new_action.docker_image, config.docker_image)
        assert_equal(
            new_action.docker_parameters,
            [{
                'key': 'test',
                'value': 123
            }],
        )
        assert_equal(new_action.env, config.env)
        assert_equal(
            new_action.extra_volumes,
            [{
                'container_path': '/nail/tmp',
                'host_path': '/tmp',
                'mode': 'RO'
            }],
        )
        assert_equal(new_action.mesos_address, config.mesos_address)

    def test_from_config_none_values(self):
        config = ConfigAction(
            name="ted",
            command="do something",
            node="first",
            executor="ssh",
        )
        new_action = action.Action.from_config(config)
        assert_equal(new_action.name, config.name)
        assert_equal(new_action.command, config.command)
        assert_equal(new_action.required_actions, [])
        assert_equal(new_action.executor, config.executor)
        assert_equal(new_action.constraints, [])
        assert_equal(new_action.docker_image, None)
        assert_equal(new_action.docker_parameters, [])
        assert_equal(new_action.env, {})
        assert_equal(new_action.extra_volumes, [])

    def test__eq__(self):
        new_action = action.Action(
            self.action.name,
            self.action.command,
            self.node_pool,
        )
        assert_equal(new_action, self.action)


if __name__ == '__main__':
    run()
